@@ -452,6 +452,7 @@ def main():
         'duk_harray.h',
         'duk_hboundfunc.h',
         'duk_hbuffer_alloc.c',
+        'duk_hbuffer_assert.c',
         'duk_hbuffer.h',
         'duk_hbuffer_ops.c',
         'duk_hbufobj.h',
@@ -461,6 +462,7 @@ def main():
         'duk_heap.h',
         'duk_heap_hashstring.c',
         'duk_heaphdr.h',
+        'duk_heaphdr_assert.c',
         'duk_heap_finalize.c',
         'duk_heap_markandsweep.c',
         'duk_heap_memory.c',
@@ -470,6 +472,7 @@ def main():
         'duk_heap_stringtable.c',
         'duk_hnatfunc.h',
         'duk_hobject_alloc.c',
+        'duk_hobject_assert.c',
         'duk_hobject_class.c',
         'duk_hobject_enum.c',
         'duk_hobject.h',
@@ -478,6 +481,7 @@ def main():
         'duk_hobject_props.c',
         'duk_hproxy.h',
         'duk_hstring.h',
+        'duk_hstring_assert.c',
         'duk_hstring_misc.c',
         'duk_hthread_alloc.c',
         'duk_hthread_builtins.c',
